# This line of code will allow shorter imports
from drukarnia_api.author import Author
from drukarnia_api.article import Article
from drukarnia_api.tag import Tag
from drukarnia_api.search import Search

from drukarnia_api.drukarnia_base import DrukarniaAPIError, DrukarniaElementDataError

print("Слава Україні 🇺🇦")
